>import Formula
>import Parser

Build automaton corresponding to formula

>type StateMap = [(Int,Formulas)]
>type TransRel = [(Int,Int)]

Build up state map and transition relation.  Arguments:

- states :: StateMap : each (n,fs) pair indicates that state number n must
  satisfy formulas fs;

- delta :: TransRel : each (m,n) pair indicates that there should be
  transitions from state m to state n;

- next : Int : the number of the next state;

- unp : [(State,Formulas)] : each (n, x) pair represents that from state
  number n, the next states must satisfy x; this represents that transitions
  from n still have to be built.

>iter :: StateMap -> TransRel -> Int -> [(Int,Formulas)] ->
>          (StateMap,TransRel)
>iter states delta next [] = (states,delta)
>iter states delta next ((n,x):unp) =
>  let satx = saturate x
>      new = -- new lists of formulas for which states have to be built
>        filter (`notElem` (map snd states)) satx
>      states' = states ++ zip [next..] new
>      unp' = unp ++ zip [next..] (map nexts new)
>      findStateNo q = head [m | (m,q') <- states', q==q']
>      delta' = delta ++ [(n,findStateNo q) | q <- satx]
>  in iter states' delta' (next+length new) unp'

>-- label :: [a] -> Int -> [(Int,a)]
>-- label xs n = zip [n..] xs

Extract "Next" formulas from list

>nexts :: Formulas -> Formulas
>nexts fs = [phi | Next phi <- fs]

Build the main automaton

>buildAuto :: Formula -> (Int, (StateMap,TransRel))
>buildAuto phi = 
>  let inits = saturate [phi]
>      states = zip [0..] inits
>      unp = zip [0..] (map nexts inits)
>      next = length states
>  in (next, iter states [] next unp)


----------------------------------------------------------------------
Make CSP specification corresponding to formula

>makeSpec0 :: Formula -> String
>makeSpec0 phi =
>  let (n, (states, delta)) = buildAuto phi
>  in -- concat (map (\ s -> show s++"\n") states) ++ "\n" ++
>     -- concat (map (\ s -> show s++"\n") delta) ++ "\n" ++
>     "SPEC =\n"++
>     if states==[] then "  div"
>     else
>       "  let\n"++concat (map (showState delta) states)++"  within "++
>       makeIntchoice ["SPEC_"++show m | m <- [0..n-1]]
>       -- makeIntchoice ["SPEC_"++show n | (n,fs) <- states, phi `elem` fs]

Show a single state (i.e. subsidiary process) of the spec

>showState delta (n,fs) =
>  let nexts = ["SPEC_"++show m | (n',m) <- delta, n==n']
>      nextsString0 = makeIntchoice nexts
>      nextsString = 
>        if length nexts<=1 then nextsString0 else "("++nextsString0++")"
>  in "    SPEC_"++show n++" = \n      " ++ makeTrans fs nextsString++"\n"

Produce CSP process to make transitions, as required by fs, to process
nextsString.

>makeTrans fs nextsString =
>  let doeses = [a | Does a <- fs]
>      doesnots = [a | Doesnot a <- fs]
>      avails = [a | Available a <- fs]
>  in case doeses of
>       [a] -> 
>         a++" -> "++nextsString++
>         concat (map (\ a -> " [] "++a++ " -> div") avails)
>       [] -> 
>         if Deadlocked `elem` fs then "STOP"
>         else if doesnots==[] then
>           "("++maybeString (Live `notElem` fs) "STOP |~| "++
>           "|~| x_:Events @ x_ -> "++nextsString++")"++
>           concat (map makeAvail avails)
>         else
>           "let poss = diff(Events,{"++commaConcat doesnots++"}) within\n"++
>           spaces 6++"("++maybeString (Live `notElem` fs) "STOP |~| "++
>           "if poss=={} then div else |~| x_:poss @ x_ -> "++
>           nextsString++")"++
>           concat (map makeAvail avails)
>         where makeAvail a = 
>                 "\n"++spaces 6++"[] "++a++ " -> "++
>                 if a `elem` doesnots then "div" else nextsString

>commaConcat = foldr1 (\ st st' -> st++","++st') 
>spaces n = replicate n ' '
>makeIntchoice ps = 
>  if ps==[] then "div" else foldr1 (\ st st' -> st++" |~| "++st') ps
>maybeString b st = if b then st else ""

The main function

>makeSpec :: String -> IO()
>makeSpec = putStr . makeSpec0 . parse
